/* $Id: e2_filelist.c 1433 2005-12-13 00:36:23Z tpgww $
 *
 * Portions copyright (C) 2004-2005 tooar <tooar@gmx.net>.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file GPL. If not, contact the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file src/e2_filelist.c 
 * @brief directory content liststore functions
 *
 * This file contains functions related to creation, filling and emptying a
 * liststore of directory content data, and fns for the associated model.
 * Also filtering of store content, and other related utilties.
 **/

#include "e2_filelist.h"
#include <string.h>
#include <sys/time.h>
#include <langinfo.h>
#include <dirent.h>
#include <pwd.h>
#include <time.h>
#include <grp.h>
#include "e2_option.h"
#include "e2_task.h"
#include "e2_dialog.h"

#ifdef E2_SELTXT_RECOLOR
extern GdkColor selectedtext;
#endif

#ifdef E2_FAM_DNOTIFY
#include <signal.h>
extern sigset_t dnotify_signal_set;
#endif
#ifdef E2_LISTTHREADS
extern GStaticMutex refresh_mutex;
#endif

//disable-refresh counter
gint refresh_ref;
//static gboolean case_sensitive; //local copy of option value, for faster access

// some filter menu callbacks (others are in filter-dialog files)

void e2_filelist_filter_toggle (FileView *view)
{
	gboolean filter_on = view->filename_filter.active 
		|| view->size_filter.active
		|| view->date_filter.active;
	gchar *name = (view == &app.pane1_view) ? _A(10) : _A(11);
	name = g_strconcat (name, ".", _A(20), NULL);
	gboolean filter_button = e2_toolbar_toggle_button_get_state (name);
	//filter_button = TRUE is 'normal' un-filtered state
	if (filter_on && filter_button)
		e2_toolbar_toggle_button_set_state (name, FALSE);
	else if (!filter_on && !filter_button)
		e2_toolbar_toggle_button_set_state (name, TRUE);
	g_free (name);
}

void e2_filelist_filter_dirs_cb (GtkWidget *widget, FileView *view)
{  //toggle cached flag to match the menu check button
//	view->filter_directories = gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (widget));
	view->filter_directories = !view->filter_directories;
	e2_fileview_refresh_list (view);
}

void e2_filelist_remove_filters_cb (GtkWidget *widget, FileView *view)
{
/* menu items now re-created for each popup
	//turn off the check menu items
	gulong handler = g_signal_handler_find (view->check_name, G_SIGNAL_MATCH_FUNC, 
		0, 0, NULL, e2_name_filter_dialog_create_cb, NULL);
	g_signal_handler_block (view->check_name, handler);
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (view->check_name), FALSE);
	g_signal_handler_unblock (view->check_name, handler);

	handler = g_signal_handler_find (view->check_size, G_SIGNAL_MATCH_FUNC, 
		0, 0, NULL, e2_size_filter_dialog_create_cb, NULL);
	g_signal_handler_block (view->check_size, handler);
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (view->check_size), FALSE);
	g_signal_handler_unblock (view->check_size, handler);

	handler = g_signal_handler_find (view->check_date, G_SIGNAL_MATCH_FUNC, 
		0, 0, NULL, e2_date_filter_dialog_create_cb, NULL);
	g_signal_handler_block (view->check_date, handler);
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (view->check_date), FALSE);
	g_signal_handler_unblock (view->check_date, handler);

 	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (view->check_dirs), FALSE);
*/
	// turn off local flags
	view->filename_filter.active = FALSE;
	view->size_filter.active = FALSE;
	view->date_filter.active = FALSE;
	view->filter_directories = FALSE;
	//show the results
	e2_filelist_filter_toggle (view);
	e2_fileview_refresh_list (view);
}
/**
 * @brief initialise file filter flags to show everything
 *
 * @param view data structure for view being processed
 *
 * @return
 */
void e2_filelist_initialize_filters (FileView *view)
{
	view->filter_directories = FALSE;

	view->filename_filter.active = FALSE;
	g_strlcpy (view->filename_filter.pattern, "*",
		      sizeof(view->filename_filter.pattern));
	view->filename_filter.case_sensitive = TRUE;

	view->size_filter.active = FALSE;
	view->size_filter.size = 0;
	view->size_filter.op = GT;

	view->date_filter.active = FALSE;
	view->date_filter.time = time(NULL);
	view->date_filter.time_type = ATIME;
	view->date_filter.op = GT;
}
/**
 * @brief decide whether an item should be included in the filelist, on the basis of its name
 *
 * Item name is compared with current filters
 * Item name string is localised
 * though this fn assumes ascii = FIXME
 *
 * @param view data structure for view being processed
 * @param info data structure for item being checked
 *
 * @return TRUE if item 'passes' (i.e. to be included in file list)
 */
static gboolean _e2_filelist_match_name_filter (FileView *view, FileInfo *info)
{
	gchar *s, *pattern, *free_this, *dup, *utf;
	gchar save;
	gboolean result;

	if (e2_fs_is_dir (info) && !view->filter_directories)
		 return TRUE;

	free_this = pattern = g_strdup (view->filename_filter.pattern);
	if (!view->filename_filter.case_sensitive)
	{
		pattern = g_utf8_strdown (pattern, -1);
		g_free (free_this);
		free_this = pattern;
	}
	utf = F_FILENAME_FROM_LOCALE (info->filename);
	//maybe several patterns, separated by commas
	while ((s = strchr (pattern, ',')) != NULL)	//if always ascii ',', don't need g_utf8_strchr()
	{	//check each pattern that is followed by a ','
		save = *s;
		*s = '\0';
		if (!view->filename_filter.case_sensitive)
		{
		  dup = g_utf8_strdown (utf, -1);
		  if (g_pattern_match_simple (pattern, dup))
		  {
		    g_free (free_this);
		    g_free (dup);
			F_FREE (utf);
		    return !view->filename_filter.invert_mask;
		  }
		  g_free(dup); 
		}
		else if (g_pattern_match_simple (pattern, utf))
		{
		  g_free(free_this);
		  F_FREE (utf);
		  return !view->filename_filter.invert_mask;
		}
		*s = save;	//in case the matched char is bigger than 1 byte
		NCHR(s);
		pattern = s;
	}
	//check the last (or only) pattern
	if (!view->filename_filter.case_sensitive)
	{
		dup = g_utf8_strdown (utf, -1);
		result = g_pattern_match_simple (pattern, dup);
		g_free (dup);
	}
	else
		result = g_pattern_match_simple (pattern, utf);

	g_free (free_this);
	F_FREE (utf);
	return (view->filename_filter.invert_mask ? !result : result);
}
/**
 * @brief decide whether an item should be included in the filelist, on the basis of its size
 *
 * Item statbuf.st_size is compared with curent filters
 *
 * @param view data structure for view being processed
 * @param info data structure for item being checked
 *
 * @return TRUE if item does NOT need to be filtered out
 */
static gboolean _e2_filelist_match_size_filter (FileView *view, FileInfo *info)
{
	if (e2_fs_is_dir (info) && !view->filter_directories)
		 return TRUE;

	switch (view->size_filter.op)
	{
		case GT:
		  return (info->statbuf.st_size > view->size_filter.size);
		  break;
		case LT:
		  return (info->statbuf.st_size < view->size_filter.size);
		  break;
		case EQ:
		  return (info->statbuf.st_size == view->size_filter.size);
		  break;
		default:
		  return TRUE;
	}
	// this path can never happen
	return TRUE;
}
/**
 * @brief decide whether an item should be included in the filelist, on the basis of its access time
 *
 * Item statbuf.st_atime is compared with curent filters
 *
 * @param view data structure for view being processed
 * @param info data structure for item being checked
 *
 * @return TRUE if item does NOT need to be filtered out
 */
static gboolean _e2_filelist_match_date_filter (FileView *view, FileInfo *info)
{
	if (e2_fs_is_dir (info) && !view->filter_directories)
		 return TRUE;

	switch (view->date_filter.time_type)
	{
		case ATIME:
		  if (view->date_filter.op == GT)
		    return (difftime(view->date_filter.time, info->statbuf.st_atime) < 0);
		  else
		    return (difftime(view->date_filter.time, info->statbuf.st_atime) > 0);
		  break;
		case MTIME:
		  if (view->date_filter.op == GT)
		    return (difftime(view->date_filter.time, info->statbuf.st_mtime) < 0);
		  else
		    return (difftime(view->date_filter.time, info->statbuf.st_mtime) > 0);
		  break;
		case CTIME:
		  if (view->date_filter.op == GT)
		    return (difftime(view->date_filter.time, info->statbuf.st_ctime) < 0);
		  else
		    return (difftime(view->date_filter.time, info->statbuf.st_ctime) > 0);
		  break;
		default:
		  return TRUE;
	}
	// this path can't happen
	return TRUE;
}
/* These functions must encapsulate any instructions that access the FileInfo
 * structs produced when a selection is interrogated.
 * This is because a refresh in the middle of such an operation would free
 * the FileInfo structs as a side-effect.
 */
/**
 * @brief  disable refreshing of config data and pane filelists
 *
 * Increase ref count, if == 1, stop the timer process which calls the refresh fns
 * Why is config data check bundled here ?
 *
 * @return
 */
void e2_filelist_disable_refresh (void)
{
	refresh_ref++;
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "disable refresh, refresh ref now = %d", refresh_ref);
#endif
	if (refresh_ref == 1)
	{
		if (e2_option_bool_get ("auto-refresh"))
		{
//don't cancel timer, to reduce risk that dirty-reports queue may over-fill
//			g_source_remove (cfg.filelist_timer_id);
			curr_view->blocked = TRUE;
			other_view->blocked = TRUE;
		}
		if (e2_option_bool_get ("auto-refresh-config"))
			//CHECKME just block this too?
			//(can't reasonably be many changes to the config file)
			g_source_remove (cfg.config_timer_id);
		e2_window_disable_status_update ();
	}
}
/**
 * @brief  cause polling of config data and pane filelists at specified intervals
 *
 * Polling (re)starts after the interval 
 * reference count is back to 0
 *
 * @param interval  milliseconds between polls
 *
 * @return
 */
/*void e2_filelist_set_refresh (guint interval)
{
	e2_filelist_disable_refresh ();
	e2_filelist_enable_refresh_custom (interval);
}

void e2_filelist_set_refresh_default (void)
{
	e2_filelist_disable_refresh ();
	e2_filelist_enable_refresh_custom (E2_FILESCHECK_INTERVAL);
} */
/**
 * @brief  re-enable polling of config data and pane filelists
 *
 * Polling (re)starts after the default 1-second interval, if the 
 * reference count is back to 0
 *
 * @return
 */
void e2_filelist_enable_refresh (void)
{
	e2_filelist_enable_refresh_custom (E2_FILESCHECK_INTERVAL);
}
/**
 * @brief  re-enable polling of config data and pane filelists, with a specified interval
 *
 * Decrease ref count, if == 0, restart the timers which call the refresh fns
 *  Config timer is bundled here because the fn refreshes the panes
 *
 * @param interval milliseconds between polls
 *
 * @return
 */
void e2_filelist_enable_refresh_custom (guint interval)
{
	refresh_ref--;
#ifdef E2_REFRESH_DEBUG
	printd (DEBUG, "enable refresh, refresh ref now = %d", refresh_ref);
#endif
	if (refresh_ref < 0)
	{
		printd (WARN, "The auto-refresh refresh count is < 0");
		e2_window_output_show (NULL, NULL);
		e2_output_print_error (
		_("Something is wrong with the auto-refresh.\nPlease submit a bug report"));
	}

	if (refresh_ref == 0)
	{
		if (e2_option_bool_get ("auto-refresh"))
		{
//			cfg.filelist_timer_id = g_timeout_add 
//				(interval, (GSourceFunc) e2_filelist_check_dirty, NULL);
			curr_view->blocked = FALSE;
			other_view->blocked = FALSE;
		}
		//this timer is used, even with fam backends
		if (e2_option_bool_get ("auto-refresh-config"))
			cfg.config_timer_id = g_timeout_add
				(E2_CONFIGCHECK_INTERVAL, (GSourceFunc) e2_option_check_config_files, NULL);
		e2_window_enable_status_update (-1);
	}
}
/**
 * @brief refresh filepane contents (and with FAM, config too) if the corresponding changed is detected
 *
 * Among other uses, this is the timeout function for filelists auto refresh
 * When FAM/gamin in use, it also checks for and flags (not processes)
 * config file updates, as that data comes from the same data stream.
 *
 * @param userdata data specified when the timer was initialised (NULL), or when called directly, non-NULL
 *
 * @return TRUE so the timer keeps working
 */
gboolean e2_filelist_check_dirty (gpointer userdata)
{
	static gboolean p1deferred = FALSE;
	static gboolean p2deferred = FALSE;
#ifdef E2_FAM
	static gboolean cfgdeferred = FALSE;
#endif

	gboolean p1dirty, p2dirty;
#ifdef E2_FAM
	gboolean localcfgdirty;	//use a local copy so that the extern value doesn't get cleared before it's noticed
	extern gboolean cfgdirty;
#endif

#ifdef E2_FAM
	if (userdata != NULL
		 && cfg.monitor_type != E2_MONITOR_DEFAULT)
			g_usleep (E2_FAMWAIT);	//wait for things to be noticed
#endif
	//this reports 'gone' as 'dirty', and then the list-refesh
	//function handles choosing a replacement dir
	e2_fs_FAM_check_dirty (&p1dirty, &p2dirty
#ifdef E2_FAM
	, &localcfgdirty
#endif
	);
	
	//CHECKME process active pane first, more likely to be attending to that ?

	if (p1dirty || p1deferred)
	{
		if (!app.pane1_view.completed	//not finished from last time
			|| (app.pane1_view.blocked && userdata == NULL))	//blocked auto-refresh
		{
			p1deferred = TRUE;
			printd (DEBUG, "pane 1 refresh deferred");
		}
		else
		{
			p1deferred = FALSE;
			e2_fileview_refresh_list (&app.pane1_view);
			printd (DEBUG, "pane 1 refreshed");
		}
	}

	if (p2dirty || p2deferred)
	{
		if (!app.pane2_view.completed
			|| (app.pane2_view.blocked && userdata == NULL))
		{
			p2deferred = TRUE;
			printd (DEBUG, "pane 2 refresh deferred");
		}
		else
		{
			p2deferred = FALSE;
			e2_fileview_refresh_list (&app.pane2_view);
			printd (DEBUG, "pane 2 refreshed");
		}
	}

#ifdef E2_FAM
	if (localcfgdirty || cfgdeferred)
	{
		if (!cfg.completed)
		{
			cfgdeferred = TRUE;
			printd (DEBUG, "config refresh deferred");
		}
		else
		{
			cfgdeferred = FALSE;
			if (localcfgdirty)
			{
				cfgdirty = TRUE;	//set, but don't clear, the main flag
				printd (DEBUG, "config refresh flag set");
			}
		}
	}
#endif
	return TRUE;
}
#ifdef E2_NEWSTORES
/**
 * @brief idle function to clear old liststores
 *
 * @param user_data UNUSED data specified when the idle was setup
 *
 * @return FALSE, to stop the callbacks
 */
/* this is for deferred liststore clearing
static gboolean _e2_filelist_clear_old_stores (gpointer user_data)
{
	GList *tmp;
	gint debug = g_list_length (app.dirty_stores);
	for (tmp	= app.dirty_stores; tmp != NULL; tmp = tmp->next)
	{
		GtkListStore *store = tmp->data;
		GtkTreeModel *mdl = GTK_TREE_MODEL (store);
		GtkTreeIter iter;
		if (gtk_tree_model_get_iter_first (mdl, &iter))
		{	//it's not empty already
			//clear file info data structs referred to in the store
			//CHECKME need to clear anything else?
			FileInfo *info;
			do
			{
				gtk_tree_model_get (mdl, &iter, FINFO, &info, -1);
				g_free (info);
			} while (gtk_tree_model_iter_next (mdl, &iter));
//			gtk_list_store_clear (store);
		}
/ *		if (tmp == app.dirty_stores //the 1st entry
			&& app.spare_store == NULL)
			app.spare_store = store;	//keep one spare store handy
		else * /
			//this does not release the corresponding memory 
			//allocated to e2, but at least that memory can be re-used
			//for subsequent stores
			g_object_unref (store);	//kill any others
	}
	g_list_free (app.dirty_stores);
	app.dirty_stores = NULL;
	printd (DEBUG, "%d old listore(s) cleared", debug);
	return FALSE;
} */
#endif
/**
 * @brief re-populate list store with unsorted data for a specified directory
 *
 * Only called from e2_fileview_prepare_list ()
 * Directory whose content we want is named in view->dir (utf-8)
 * Current store content is cleared after successful preliminary things,
 * but before new data is added, of course
 * Active filter(s) are applied, but no sorting.
 * If relevant, the text-color for selected items is updated here, rather 
 * than in the selection function (too many un-necessary repeats, there)
 * This function is thread-safe
 *
 * @param view data structure for the view to which the file list belongs
 *
 * @return
 */
#ifdef E2_NEWSTORES
GtkListStore *
#else
void
#endif
	e2_filelist_refill_store (FileView *view)
{
//	printd (DEBUG, "start store refill");
	//FIXME adapt these for vfs
	DIR *dp;
	struct dirent *entry;
#ifndef E2_FASTREAD
	struct dirent entrybuf;
#endif
	FileInfo info;

	gchar *loc_dir = F_FILENAME_TO_LOCALE (view->dir);

	if (e2_fs_stat (loc_dir, &info.statbuf))	//stat before opendir
	{
		printd (WARN, "Unable to stat directory: %s", view->dir);
		F_FREE (loc_dir);
#ifdef E2_NEWSTORES
		return NULL;
#else
		return;
#endif
	}

	dp = opendir (loc_dir);	//FIXME vfs
	F_FREE (loc_dir);

/*	would be nice to prevent access-time updates purely due to
	refreshing, but the atime really does change as a result
	of the refresh process, and if we were to successfully revert
	it, the ctime would be changed instead
	times are immediately shown in the other pane if it's the
	parent of this one
	if we could suspend monitoring of such parent dir, we
	must not lose any prior reports for there
	QUERY can the changed atime itself trigger a report which
	causes a subsequent refresh, creating an endless cycle?
	especially with multiple threads and/or processors ?
	Assuming it can do so, we clear the reports 'queue'
	after the dir is opened
	Would be nice if we could just clear any report(s) due to
	the refresh per se - but the reporting is not that detailed */
/*#if defined (E2_FAM_INOTIFY) || defined (E2_FAM_DNOTIFY) || defined (E2_FAM_KQUEUE)
	if (view->refresh)
		e2_fs_FAM_cancel_monitor_dir (view->dir);
#endif */

#if defined (E2_FAM_INOTIFY) || defined (E2_FAM_DNOTIFY) || defined (E2_FAM_KQUEUE)
	if (view->refresh)
		e2_fs_FAM_clean_reports (view->dir);
/*	prior reports about the dir are now gone.
	Henceforth any report about the dir or an item in it
	will trigger a new refresh, even if the change that
	caused the report is shown in this refresh.
	The alternative is to do this at the end, and lose any
	change after an item is processed in this refresh */
#endif	//which fam

	if (dp == NULL)
	{
		printd (WARN, "Unable to open directory: %s", view->dir);
#ifdef E2_NEWSTORES
		return NULL;
#else
		return;
#endif
	}

/*#if defined (E2_FAM_DNOTIFY)
	//cut down on spurious context changes
	//block DNOTIFY_SIGNAL
	sigprocmask (SIG_BLOCK, &dnotify_signal_set, NULL);
#endif */

#ifdef E2_FASTREAD
	//quickly grab a copy of all dirents for the dir
	GList *entries = NULL;
	struct dirent *entry2;
	E2_LOCAL_LOCK
	while (TRUE)
	{
		entry = g_malloc (sizeof (struct dirent));
		if (readdir_r (dp, entry, &entry2) || entry2 == NULL)	//FIXME vfs
		{
			g_free (entry);
			break;
		}
		entries = g_list_append (entries, entry);
	}
	closedir (dp);
	E2_LOCAL_UNLOCK
#if defined (E2_FAM_INOTIFY) || defined (E2_FAM_DNOTIFY) || defined (E2_FAM_KQUEUE)
//	if (view->refresh)
//		e2_fs_FAM_clean_reports (view->dir);
//	if (view->refresh)	FIXME adjust reports for other pane if relevant
//		e2_fs_FAM_clean_reports (rt->path);
#endif	//which fam
#endif	//def E2_FASTREAD

//#ifndef E2_FAM
	//these are not really necessary when fam is working,
	//but any time saving by testing for that is not material ..
	view->dir_mtime = info.statbuf.st_mtime;
	view->dir_ctime = info.statbuf.st_ctime;
//#endif

	E2_LOCK
#ifdef E2_NEWSTORES
/*	//arrange to cleanup the old store at an idle time
	//prevent old store from being zapped when the model is changed
	g_object_ref (view->store);
	gboolean newtimer = (app.dirty_stores == NULL);
	app.dirty_stores = g_list_append (app.dirty_stores, view->store);
	if (newtimer)
	{
		printd (DEBUG, "setup to clear stores later");
		g_idle_add (_e2_filelist_clear_old_stores, NULL);
	} */
//	if (app.spare_store != NULL)
//		printd (DEBUG, "re-using old backup store");
	//use a re-cycled liststore if we can
	GtkListStore *store = 
//	(app.spare_store != NULL) ?
//		app.spare_store : 
	e2_filelist_make_store ();
#else
	GtkListStore *store = view->store;
#endif	//def E2_NEWSTORES
	//now we're good to go, clear the current store, in preparation for refilling
	GtkTreeModel *mdl = GTK_TREE_MODEL (view->store);
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter_first (mdl, &iter))
	{	//it's not empty already
		//clear file info data structs referred to in the store
		//CHECKME need to clear anything else?
		FileInfo *info;
		do
		{
			gtk_tree_model_get (mdl, &iter, FINFO, &info, -1);
			g_free (info);
		} while (gtk_tree_model_iter_next (mdl, &iter));
		gtk_list_store_clear (view->store);
	}

	E2_UNLOCK

	//get format for dates display
	gchar *strf_string;
	switch	(e2_option_int_get ("date-string"))
	{
		case 1:
			strf_string = "%d/%m/%y %H:%M";	//standard
			break;
		case 2:
			strf_string = "%m/%d/%y %H:%M";	//american
			break;
		case 3:
/*			{
				strf_string = nl_langinfo (D_T_FMT);
				if (strf_string != NULL && *strf_string != '\0')
					break;
			} */
			strf_string = "%x %X";	//localised
			break;
		default:
			strf_string = "%b %d %H:%M";
			break;
	}
	//get format for size display
	gchar *comma;
	switch	(e2_option_int_get ("size-string"))	
	{
		case 1:
		{
			comma = nl_langinfo (THOUSEP);
			if (comma == NULL || *comma == '\0')
				comma = ",";
			break;
		}
		default:
			comma = NULL;	//signal to use the condensed version
			break;
	}
	
	gboolean caseignore = ! e2_option_bool_get ("namesort-case-sensitive");

	E2_LOCK
	GtkStyle *style = gtk_rc_get_style (view->treeview);
	E2_UNLOCK

	GdkColor *default_color = &style->text[GTK_STATE_NORMAL];
//	GdkColor *background = &style->base[GTK_STATE_NORMAL];
#ifdef E2_SELTXT_RECOLOR
	// set the current foreground color for selected items
	//used in selection function, but don't want to do this each selection
	//CHECKME get the right one - ~bg[*] is no good nor is ~base[*] 
	//~text is likely to be a light color
	selectedtext.red = ~style->text[GTK_STATE_NORMAL].red;
	selectedtext.green = ~style->text[GTK_STATE_NORMAL].green;
	selectedtext.blue = ~style->text[GTK_STATE_NORMAL].blue;
#endif

	FileInfo *this_info;
	struct tm *tm_ptr;
	struct passwd *pwd_buf;
	struct group *grp_buf;
	gchar filename_buf[sizeof(this_info->filename)];
	gchar size_buf[20];	//enough for 999 Tb
	gchar modified_buf[25];
	gchar accessed_buf[25];
	gchar changed_buf[25];
	gchar perm_buf[11];
	gchar uid_buf[20];
	gchar gid_buf[20];
	gchar *buf[NAMEKEY+1];	//pointers to NAMEKEY strings

/* seems like setting up a GValue array is not worth it ...
#if defined (E2_CURRENTGTK_VERSION) && GTK_MINOR_VERSION >= 6
#ifdef E2_SELTXT_RECOLOR
#define STORECOLCOUNT MODEL_COLUMNS
#else
//BACKCOLOR info not set in store when it's filled, so not needed
//in setup arrays unless SELCOLOR is in effect
#define STORECOLCOUNT MODEL_COLUMNS-1
#endif

static gint colnums[STORECOLCOUNT] = {
		FILENAME, SIZE, PERM, OWNER, GROUP,
		MODIFIED, ACCESSED, CHANGED, NAMEKEY,
		FINFO, FORECOLOR
#ifdef E2_SELTXT_RECOLOR
	, BACKCOLOR
	, SELCOLOR
#endif
		};

	//FIXME do this init just once
	GValue *colvalues[STORECOLCOUNT];
	gint i;
	for (i=FILENAME; i <= NAMEKEY ; i++)
	{
		g_value_init (colvalues[i], G_TYPE_STRING);
		g_value_set_static_string (colvalues[i], buf[i]);
	}
	for (; i < STORECOLCOUNT; i++);
		g_value_init (colvalues[i], G_TYPE_???);

#endif
*/
	gchar *freeme;
	//look these up once, for speed
	GdkColor *link_color = e2_option_color_get ("color-ft-link");
	GdkColor *dir_color = e2_option_color_get ("color-ft-dir");
	GdkColor *dev_color = e2_option_color_get ("color-ft-dev");
	GdkColor *sock_color = e2_option_color_get ("color-ft-socket");
	GdkColor *exec_color = e2_option_color_get ("color-ft-exec");
	gboolean test1 = !e2_option_bool_get ("show-updir-entry")
			|| g_str_equal (view->dir, G_DIR_SEPARATOR_S);  //at root directory now?

	//change current dir coz statted items in the dir have no path
	//no success-check, we've already done 2, above
	e2_fs_chdir (view->dir);

	//iterate through the directory contents
#ifdef E2_FASTREAD
	GList *tmp;
	for (tmp = entries; tmp != NULL; tmp = tmp->next)
	{
		entry = tmp->data;
#else
	while (!readdir_r (dp, &entrybuf, &entry) && entry != NULL)	//FIXME vfs
	{
#endif	//def E2_FASTREAD
		//some items we're not interested in
		if (entry->d_name[0] == '.'
			&& (
					(entry->d_name[1] == '\0')
				|| (test1 && g_str_equal (entry->d_name, ".."))
				|| (!g_str_equal (entry->d_name, "..") && !view->show_hidden)
					)
			)
		{
#ifdef E2_FASTREAD
				g_free (entry);
#endif
				continue;
		}

		//get data into local structure for filter tests
		//(any truncation means it will never be able to be stat'd anymore)
		g_strlcpy (info.filename, entry->d_name, sizeof (info.filename));
		e2_fs_lstat (entry->d_name, &info.statbuf);
#ifdef E2_FASTREAD
		g_free (entry);
#endif
		if (view->filename_filter.active && !_e2_filelist_match_name_filter (view, &info))
		    continue;
		if (view->size_filter.active && !_e2_filelist_match_size_filter (view, &info))
		    continue;
		if (view->date_filter.active && !_e2_filelist_match_date_filter (view, &info))
		    continue;

		//this one is to keep
		this_info = g_memdup (&info, sizeof(FileInfo));
		//localised string, no glib copying
		//name is already truncated to buffer size, if need be, so no overflow here
		strcpy (filename_buf, info.filename);

		//directories (and links to dirs) get a trailing separator, if there's space
		if (e2_fs_is_dir (this_info))
		{
			guint len = strlen (filename_buf);
			if (len < sizeof(filename_buf) -1)
			{
				*(filename_buf+len) = G_DIR_SEPARATOR;
				*(filename_buf+len+1) = '\0';
			}
		}
		//convert to utf if appropriate
		buf[FILENAME] = D_FILENAME_FROM_LOCALE (filename_buf);

		if (comma == NULL)
		{	//use condensed size format
			if (info.statbuf.st_size < 10240) // less than 10k
			{
			  g_snprintf(size_buf, sizeof(size_buf), "%llu",
					info.statbuf.st_size);
			}
			else if (info.statbuf.st_size < 1048576) // less than a meg
			{
			  g_snprintf(size_buf, sizeof(size_buf), "%.1f%s",
				(gfloat) info.statbuf.st_size / 1024.0, _("k"));
			}
			else  // more than a meg
			{
			  g_snprintf(size_buf, sizeof(size_buf), "%.1f%s",
				(gfloat) info.statbuf.st_size / 1048576.0, _("M"));
			}
		}
		else
		{	//use actual size, with commas
			g_snprintf(size_buf, sizeof(size_buf), "%llu",
				info.statbuf.st_size);

			guint len = strlen (size_buf);
			guint ths = len-1;  //0-based index
			guint i;
			while (ths > 2 && len < sizeof(size_buf))
			{
				for (i = len-1; i > ths-3; i--)
					size_buf[i+1] = size_buf[i];

				size_buf[i+1] = *comma;
				size_buf[++len] = '\0';
				ths = i;
			}
		}
		buf[SIZE] = g_strdup (size_buf);	//content is ascii, no need to convert to utf

		e2_fs_get_perm_string (perm_buf, sizeof(perm_buf), info.statbuf.st_mode);
//		buf[PERM] = e2_utf8_from_locale (perm_buf);
		buf[PERM] = g_strdup (perm_buf); //content is ascii, no need to convert to utf

		if ((pwd_buf = getpwuid (info.statbuf.st_uid)) == NULL)
		{
		  g_snprintf (uid_buf, sizeof(uid_buf), "%d",
		     (guint) info.statbuf.st_uid);
		  buf[OWNER] = g_strdup (uid_buf); //content is ascii number, no need to convert to utf
		}
		else
		{
		  buf[OWNER] = e2_utf8_from_locale (pwd_buf->pw_name);
		}

		if ((grp_buf = getgrgid (info.statbuf.st_gid)) == NULL)
		{
		  g_snprintf(gid_buf, sizeof(gid_buf), "%d",
		       (guint) info.statbuf.st_gid);
		  buf[GROUP] = g_strdup (gid_buf); //content is ascii number, no need to convert to utf
		}
		else
		{
		  buf[GROUP] = e2_utf8_from_locale (grp_buf->gr_name);
		}
		tm_ptr = localtime (&(info.statbuf.st_mtime));
		strftime (modified_buf, sizeof(modified_buf), strf_string, tm_ptr);
		buf[MODIFIED] = e2_utf8_from_locale (modified_buf);

		tm_ptr = localtime (&(info.statbuf.st_atime));
		strftime (accessed_buf, sizeof(accessed_buf), strf_string, tm_ptr);
		buf[ACCESSED] = e2_utf8_from_locale (accessed_buf);

		tm_ptr = localtime(&(info.statbuf.st_ctime));
		strftime (changed_buf, sizeof(changed_buf), strf_string, tm_ptr);
		buf[CHANGED] = e2_utf8_from_locale (changed_buf);

		if (caseignore)
		{
			freeme = g_utf8_casefold (buf[FILENAME], -1);
#if  !defined (E2_CURRENTGTK_VERSION) || GTK_MINOR_VERSION < 8
			buf[NAMEKEY] = g_utf8_collate_key (freeme, -1);
#else
			buf[NAMEKEY] = g_utf8_collate_key_for_filename (freeme, -1);
#endif
			g_free (freeme);
		}
		else
#if  !defined (E2_CURRENTGTK_VERSION) || GTK_MINOR_VERSION < 8
			buf[NAMEKEY] = g_utf8_collate_key (buf[FILENAME], -1);
#else
			buf[NAMEKEY] = g_utf8_collate_key_for_filename (buf[FILENAME], -1);
#endif
		GdkColor *foreground;
		switch (this_info->statbuf.st_mode & S_IFMT)
		{
		  case S_IFLNK:
		    foreground = link_color;
		    break;
		  case S_IFDIR:
		    foreground = dir_color;
		    break;
		  case S_IFCHR:
		  case S_IFBLK:
		    foreground = dev_color;
		    break;
		  case S_IFSOCK:
		    foreground = sock_color;
		    break;
		  case S_IFREG:
			  //show as executable if _anyone_ has that permission
		    if ((S_IXUSR & this_info->statbuf.st_mode)
		  		|| (S_IXGRP & this_info->statbuf.st_mode)
		  		|| (S_IXOTH & this_info->statbuf.st_mode))
			  foreground = exec_color;
			else
			  foreground = default_color;
		    break;
		  default:
			foreground = default_color;
		    break;
		}

		E2_LOCK

		GtkTreeIter iter;
#if !defined (E2_CURRENTGTK_VERSION) || GTK_MINOR_VERSION < 6
		//append empty row to the list store
		gtk_list_store_append (store, &iter);

//		gettimeofday (&strt, NULL); //FOR DEBUGGING, check time this load takes
		
		//populate it
		gtk_list_store_set (store, &iter,
#else
		gtk_list_store_insert_with_values (store, &iter, -1,
#endif
			FILENAME, buf[FILENAME],
			SIZE, buf[SIZE],
			PERM, buf[PERM],
			OWNER, buf[OWNER],
			GROUP, buf[GROUP],
			MODIFIED, buf[MODIFIED],
			ACCESSED, buf[ACCESSED],
			CHANGED, buf[CHANGED],
			NAMEKEY, buf[NAMEKEY],
			FINFO, this_info,
			FORECOLOR, foreground,
//			BACKCOLOR, background,	don't set this till needed
#ifdef E2_SELTXT_RECOLOR
			SELCOLOR, foreground,  //for reverting when item unselected
#endif
//			VISIBLE, TRUE,
			-1);
//		gettimeofday (&nd, NULL);

		E2_UNLOCK

		//cleanup
		gint i;
		for (i = 0; i <= NAMEKEY; i++)
			g_free (buf[i]);

/*		if (nd.tv_usec < strt.tv_usec)
		{
		    int nsec = (strt.tv_usec - nd.tv_usec) / 1000000 + 1;
    		strt.tv_usec -= 1000000 * nsec;
    		strt.tv_sec += nsec;
		}
		if (nd.tv_usec - strt.tv_usec > 1000000)
		{
			int nsec = (strt.tv_usec - nd.tv_usec) / 1000000;
			strt.tv_usec += 1000000 * nsec;
			strt.tv_sec -= nsec;
		}
		result.tv_sec = result.tv_sec + nd.tv_sec - strt.tv_sec;
		result.tv_usec = result.tv_usec + nd.tv_usec - strt.tv_usec;
		if (result.tv_usec > 1000000)
		{
			int nsec = result.tv_usec / 1000000 + 1;
    		result.tv_usec -= 1000000 * nsec;
    		result.tv_sec += nsec;
		}
*/
	}

#ifdef E2_FASTREAD
	g_list_free (entries);
#else
	closedir (dp);	//FIXME vfs
#endif	//def E2_FASTREAD

#ifdef E2_LISTTHREADS
	printd (DEBUG, "finish store refill for %s", view->dir);
#endif
//	printd (DEBUG, "populating rows took %f seconds", result.tv_sec + result.tv_usec / 1000000.0 );
/*#ifdef E2_FAM_DNOTIFY
	//unblock DNOTIFY_SIGNAL
	sigprocmask (SIG_UNBLOCK, &dnotify_signal_set, NULL);
#endif */
#ifdef E2_NEWSTORES
	return store;
#endif
}
/**
 * @brief create list store structure
 *
 * Creates basic structure of list store, for panes filelists, and stores
 * pointer to the store in the view rt data.
 * No rows or data are added.
 * In any event, it destroys any existing list-store belonging to the view (superfluous ?)
 *
 * @return the liststore
 */
GtkListStore *e2_filelist_make_store (void)
{
//	if (GTK_LIST_STORE (view->store) != NULL)
//		gtk_widget_destroy (GTK_LIST_STORE(view->store));   how to destroy old version ?
//	view->store
	GtkListStore *store = gtk_list_store_new (MODEL_COLUMNS,
	G_TYPE_STRING,  //FILENAME
	G_TYPE_STRING,  //SIZE
	G_TYPE_STRING,  //PERM
	G_TYPE_STRING,  //OWNER
	G_TYPE_STRING,  //GROUP
	G_TYPE_STRING,  //MODIFIED
	G_TYPE_STRING,  //ACCESSED
	G_TYPE_STRING,  //CHANGED
	// the rest are not displayed
	G_TYPE_STRING,  //NAMEKEY for i18n name sorts
	G_TYPE_POINTER,  //FINFO pr to FileInfo for the item
	GDK_TYPE_COLOR,  //FORECOLOR line colour
	GDK_TYPE_COLOR  //BACKCOLOR line colour
#ifdef E2_SELTXT_RECOLOR
	, GDK_TYPE_COLOR  //SELCOLOR  foreground of selected lines
#endif
//	G_TYPE_BOOLEAN	//VISIBLE whether filtered or not
	);
	return store;
}
